#include <CCircuit.hh>

#include <electro/CResistor.hh>
#include <electro/CCapasistor.hh>
#include <electro/CInductor.hh>
#include <electro/CVoltage.hh>
#include <electro/CCurent.hh>
#include <electro/CCurent_u.hh>
#include <electro/CCurent_i.hh>
#include <electro/CVoltage_u.hh>
#include <electro/CVoltage_i.hh>


#include <xml/Parser.hh>
#include <xml/Tokenizer.hh>

#include <math/Complex.hh>
#include <math/CMatrix.hh>
#include <math/CConversion.hh>

#include <networksarea.hh>
#include <CRlcException.hh>


const std::string _RLC_MAIN_TAG="rollercoaster";
const std::string _RLC_ELEMENTS_TAG="elements";
const std::string _RLC_NETS_TAG="nets";
const std::string _RLC_ELEMENTS_RESISTOR="res";
const std::string _RLC_ELEMENTS_CAPASISTOR="cap";
const std::string _RLC_ELEMENTS_INDUCTOR="ind";
const std::string _RLC_ELEMENTS_VOLTAGE="vol";
const std::string _RLC_ELEMENTS_CURENT="cur";
const std::string _RLC_ELEMENTS_CURENT_U="cur_u";
const std::string _RLC_ELEMENTS_CURENT_I="cur_i";
const std::string _RLC_ELEMENTS_VOLTAGE_U="vol_u";
const std::string _RLC_ELEMENTS_VOLTAGE_I="vol_i";
const std::string _RLC_ELEMENTS_NODE="node";
const std::string _RLC_NETS="net";
    
    


/*! **************************************************************************/
CElement * CCircuit::get_item(std::string s)
{
   for(int i=0;i<items.size();++i)	
      if((*items[i]).getName()==s) return items[i];
   return NULL;
}

CCircuit::~CCircuit(){
 for(int i=0;i<items.size();i++)
   delete items[i];
}

/*! ***************************************************************************/
void CCircuit::conect_items(std::string nameA,unsigned int pinA,std::string nameB,unsigned int pinB)
{
  CElement * elA=get_item(nameA);
  CElement * elB=get_item(nameB);
  
  if(!elA) throwError("Nie znaleziono elementu "+nameA);
  if(!elB) throwError("Nie znaleziono elementu "+nameB);

  if(!elA->conect(elB,pinA,pinB)) throwError("Nie poprawny pin dla elementu "+nameA);
  if(!elB->conect(elA,pinB,pinA)) throwError("Nie poprawny pin dla elementu "+nameB);
}

/*! ***************************************************************************/
void CCircuit::throwError(const string& inMessage) const
{
	throw CRlcException("CShema::"+inMessage);
}

/*! *************************************************************************/
void CCircuit::load(const std::string & we)
{
    //ustalamy plik na ktorym pracujemy
    XML::Parser parser;    
    filein = we;
	
    //probujemy go sparsowac
    try
    {
    	parser.parse(filein);
    }
    catch( std::exception& e)
    {
        //jesli sie nie udalo to wywalamy blad jaki zwrocil paser
	throwError( e.what());
    }
    catch(...)
    {
      cerr<<"Nie znany wyjatek\n";
    }

    XML::Parser::Node * n = parser.getRootNode();
    XML::Parser::Node * m = NULL;

    //sprawdzamy czy nasze drzewo nie jest puste
    if (!n) throwError("Drzewo puste");
	
    //jesli nie to szukamy znacznika glownego elektronika
    do
    {
	if(n->getValue()==_RLC_MAIN_TAG) break;
    }while(n=n->getNextSibling());    
    

    
    // sprawdzamy czy takowy znajduje sie gdzies
    if (!n) throwError("Brak znacznika "+_RLC_MAIN_TAG);
	
    //jesli sie znajduje to bierzemy pierwsze dziecko
    //w m trzymamy pierwszego potomka
    m=n->getFirstChild();
	
    //szukamy elementow
    n=m;
    while(n && n->getValue()!=_RLC_ELEMENTS_TAG)
		n=n->getNextSibling();

    if(n) 
       {
            //wczytujemy sobie polaczenia
            std::string type,name;
            int x,y,rot;
            XML::Parser::Node * r= n->getFirstChild();
            if(r) //jesli sa jakies dzieci
            do
            {
                type=r->getValue();
                name=r->getTagAttribute("name");
                x=StrToInt(r->getTagAttribute("x").c_str(),0);
                y=StrToInt(r->getTagAttribute("y").c_str(),0);
                rot=StrToInt(r->getTagAttribute("rot").c_str(),0);
                
                if(get_item(name)) throwError("Element o nazwie "+ name + " juz istnieje.");
                
                try{
                  
               	if (type==_RLC_ELEMENTS_RESISTOR){
         	        float R=StrToFloat(r->getTagAttribute("r"),1);
           	        add_item(new CResistor(name,R,x,y,rot));
            		}
                else if (type==_RLC_ELEMENTS_CAPASISTOR){
            	    	float c=StrToFloat(r->getTagAttribute("c"),1);
            	    	add_item(new CCapasistor(name,c,x,y,rot));
            		}
                else if (type==_RLC_ELEMENTS_INDUCTOR){
            	    	float l=StrToFloat(r->getTagAttribute("l"),1);
            	    	add_item(new CInductor(name,l,x,y,rot));
            		}
                else if (type==_RLC_ELEMENTS_VOLTAGE){
            	    	float v=StrToFloat(r->getTagAttribute("v"),1);
                        double omega;
                        if(r->getTagAttribute("omega")!="")
                          omega=StrToFloat(r->getTagAttribute("omega"),0);
                        else if(r->getTagAttribute("f")!="")  
                          omega=StrToFloat(r->getTagAttribute("f"),0)*2*M_PI;
                        else
                          omega=0;  
                        double fi0=StrToFloat(r->getTagAttribute("fi0"),0);
            	    	add_item(new CVoltage(name,v,omega,fi0,x,y,rot));
            		}
                else if (type==_RLC_ELEMENTS_CURENT){
            	    	float i=StrToFloat(r->getTagAttribute("i"),1);
                        double omega;
                        if(r->getTagAttribute("omega")!="")
                          omega=StrToFloat(r->getTagAttribute("omega"),0);
                        else if(r->getTagAttribute("f")!="")  
                          omega=StrToFloat(r->getTagAttribute("f"),0)*2*M_PI;
                        else
                          omega=0;  
                        double fi0=StrToFloat(r->getTagAttribute("fi0"),0);
            	    	add_item(new CCurent(name,i,omega,fi0,x,y,rot));
            		}
                else if (type==_RLC_ELEMENTS_CURENT_U){
                        float gm=StrToFloat(r->getTagAttribute("gm"),1);
                        std::string on=r->getTagAttribute("on");
                        add_item(new CCurent_u(name,gm,on,x,y,rot));
                        }
                else if (type==_RLC_ELEMENTS_CURENT_I){
                        float gm=StrToFloat(r->getTagAttribute("gm"),1);
                        std::string on=r->getTagAttribute("on");
                        add_item(new CCurent_i(name,gm,on,x,y,rot));
                        }
                else if (type==_RLC_ELEMENTS_VOLTAGE_I){
                        float gm=StrToFloat(r->getTagAttribute("gm"),1);
                        std::string on=r->getTagAttribute("on");
                        add_item(new CVoltage_i(name,gm,on,x,y,rot));
                        }
                else if (type==_RLC_ELEMENTS_VOLTAGE_U){
                        float gm=StrToFloat(r->getTagAttribute("gm"),1);
                        std::string on=r->getTagAttribute("on");
                        add_item(new CVoltage_u(name,gm,on,x,y,rot));
                        }
                else if (type==_RLC_ELEMENTS_NODE){
            	    	add_item(new CNode(name,x,y));
            		}
		else 
		        std::cerr << "Unknown tag type \""<<type<<"\""<<endl;
                }
                catch(CRlcException &e){
                    std::cerr<<e.what()<<endl;
                }
            }while(r=r->getNextSibling());
            else throwError("Brak znacznika w "+_RLC_ELEMENTS_TAG);
	}
    else throwError("Brak znacznika "+_RLC_ELEMENTS_TAG);

    //szukamy polaczen
    n=m;
    while(n && n->getValue()!=_RLC_NETS_TAG)
		n=n->getNextSibling();

    if(n) 
	{
            std::string type;
            XML::Parser::Node * r= n->getFirstChild();
            if(r) //jesli sa jakies dzieciaki
            do
            {
               	type=r->getValue();
//		cout << type << endl ; 
               	if (type==_RLC_NETS){
                    std::string nameA=r->getTagAttribute("nameA");
                    std::string nameB=r->getTagAttribute("nameB");
               	    int pinA=atoi(r->getTagAttribute("pinA").c_str());
               	    int pinB=atoi(r->getTagAttribute("pinB").c_str());
                    if(nameA!=nameB)
       		      conect_items(nameA,pinA,nameB,pinB);
                    else
                      throwError("Nie mozna dodac scierzki laczacej ten sam element.");
		}
		else 
		   std::cerr << "Unknown tag type \""<<type<<"\""<<endl;
            }while(r=r->getNextSibling());
            else throwError("Brak wpisow w "+_RLC_NETS_TAG);

	}
    else throwError("Brak znacznika "+_RLC_NETS_TAG);


    // Setting this circuit as parent of every elements, so that they know to who they must
    // turn when they want to draw themselves.
   for(int i=0;i<items.size();++i)	
      (*items[i]).setParent(this);
    

}

void CCircuit::print_items()
{
//   for(int i=0;i<items.size();++i)	
//      (*items[i]).print();
   for(int i=0;i<items.size();++i)	
      (*items[i]).print_conections();

}

void CCircuit::clear(){
//teraz musimy wywalic wszyko co sie tylko da :]

//elementow
try{
//  for(int i=0;i<items.size();i++)
//    delete items[i];
}
catch(...){
 throw CRlcException("CCircuit::clean() : Nie oczekiwany blad podczas usuwania elementow.");
}
//czyscimy liste
items.clear();
//zapominamy nazwe pliku
filein="";
};

void CCircuit::draw()
{
  m_ParentDrawingArea->m_RefGC_NetworksArea->set_foreground(m_ParentDrawingArea->m_Color_Branch);
  // std::cout << "1" << std::endl;
  for(int i=0;i<items.size();++i) {
    // std::cout << i << "=" << &(*items[i]) << "(" << items[i]->getName() << ")" << std::endl;
    // if ( !((13 <= i) && (i <= 25))  )
      items[i]->drawConnections();
  }

  // std::cout << "2" << std::endl;
  m_ParentDrawingArea->m_RefGC_NetworksArea->set_foreground(m_ParentDrawingArea->m_Color_Element);
  for(int i=0;i<items.size();++i) { 
    items[i]->draw();
  }
  // std::cout << "3" << std::endl;
  m_ParentDrawingArea->m_RefGC_NetworksArea->set_foreground(m_ParentDrawingArea->m_Color_NameText);
  for(int i=0;i<items.size();++i)  
    items[i]->drawName();

  // std::cout << "4" << std::endl;
  m_ParentDrawingArea->m_RefGC_NetworksArea->set_foreground(m_ParentDrawingArea->m_Color_ValueText);
  for(int i=0;i<items.size();++i)  
    items[i]->drawValue();

  // std::cout << "5" << std::endl;
  m_ParentDrawingArea->m_RefGC_NetworksArea->set_foreground(m_ParentDrawingArea->m_Color_FG);
}

void CCircuit::setParent(RLC::CNetworksArea* parent)
{
  m_ParentDrawingArea = parent;
}

