#ifndef CElement_HH
#define CElement_HH

#include <math/Complex.hh>
#include <electro/CSignal.hh>
#include <electro/CBranch.hh>
#include <iostream>
#include <vector>


const double  MAX_DOUBLE=1.0e99;

class CElement;
class CBranch;
class CCircuit;

  /**
   * Klasa reprezentuje pin ukadu (koncowka do ktorej cos moze byc podlaczone).
   */
  class CPin
  {
    public:
      /**
       * Konstruktor.
       * @param *dst_item wskanik do elementu docelowego.
       * @param dst_pin numer pinu w elemencie docelowym.
       * @see ~CPin();
       */
      CPin(CElement *dst_item=NULL, unsigned int dst_pin=0){element=dst_item;pin=dst_pin;}
      /**
       * Destruktor.
       * @see CPin();
       */       
      ~CPin(){pin=0;element=NULL;};
      /**
       * Wskaznik do elementu docelowego do ktorego nasza klasa jest podlaczaona.
       * Zdajemy sobie sprawe iz ten element powinien byc elementem prywatnym, jednak
       * zostal tak zaimplementowany w celu jego zrozumialszego wykorzystania w dalszym kodzie
       */       
      CElement *element;
      /**
       * Numer pinu elementu docelowego do ktrego nasza klasa jest podlaczona.
       * Zdajemy sobie sprawe iz ten element powinien byc elementem prywatnym, jednak
       * zostal tak zaimplementowany w celu jego zrozumialszego wykorzystania w dalszym kodzie
       */       
      unsigned int pin;
    };

  
  /**
   * Klasa reprezentuje pojedynczy element w ukladzie.
   */
  class CElement
  {
    public:
      /**
       * Konstruktor.
       * @param symbol typ elementu.
       * @param name nazwa.
       * @param x rzedna elementu na schemacie.
       * @param y odcieta elementu na schemacie.
       * @param rot rotacja elementu na schemacie, mozliwoci : prawo(0), gora(1), lewo(2),
       *        dol(3). Pozostale wartosc sa rownowazne z rotacja prawa.
       * @param proportion a proportion of an arrow. This value is understood as the part
       *        of squareSize that will be the half of arrow's length.
       * @see ~CElement();
       */
      CElement(std::string symbol="",std::string name="",int x=0,int y=0,int rot=0);
      
      /**
       * Destructor.
       * @see Element();
       */
      ~CElement();

      /**
       * Zwraca jednostke w postaci stringu.
       * @return jednostka m_unitSymbol elementu w postaci lancucha znakow.
       * @see m_unitSymbol
       */ 
      std::string getUnitSymbol() {return m_unitSymbol;}
      /**
       * Zwraca symbol wartosci w postaci stringu.
       * @return symbol wartosci w postaci lancucha znakow.
       * @see m_unitSymbol
       */ 
      virtual std::string getValueSymbol() {return m_valueSymbol;}
      /**
       * Zwraca wartosc elementu w postaci stringu.
       * @return wartosc razem z jednostkami. Dla sterowanych odpowiedni string.
       * @see getValue
       */ 
      virtual std::string getValueString();
      /**
       * Zwarca wartosc elementu.
       * Dla rezystora opornosc R, dla cewki indukcje L, dla kondensatora 
       * pojemnosc C, etc. Jesli nie mozna zwrocic wartosci (np nie ma czego
       * zwaracac) zwracane jest zero.
       * @return wartosc wielkosci charakteryzujacej element
       */ 
      virtual double getValue() {return 0;}
      
      
      /**
       * Zwraca typ elementu.
       * @return typ elementu
       */ 
      std::string getType(void){return m_symbol;}
      /**
       * Zwraca nazwe elementu.
       * @return nazwa elementu
       * @see setName();        
       */ 
      std::string getName(void){return m_name;}
      /**
       * Zwraca rzedna elementu.
       * @return rzedna elementu
       * @see getY(); getRot(); setX(); setY(); setRot();
       */ 
      int getX(void){return m_x;}
      /**
       * Zwraca odcieta elementu.
       * @return odcieta elementu
       * @see getX(); getRot(); setX(); setY(); setRot();
       */ 
      int getY(void){return m_y;}
      /**
       * Zwraca rotacje elementu.
       * @return rotacja elementu
       * @see getY(); getX(); setX(); setY(); setRot();
       */ 
      int getRot(void){return m_rot;}
    
      /**
       * Ustawia nazwe elementu.
       * @param name nowa nazwa
       * @see getName(); 
       */ 
      void setName(std::string name){m_name=name;}
//      void setSymbol(std::string symbol){m_symbol=symbol;}
      /**
       * Ustawia rzedna elementu na schemacie.
       * @param x nowa rzedna
       * @see getX(); getY(); getRot(); setY(); setRot();
       */ 
      void setX(int x){m_x=x;}
      /**
       * Ustawia odcieta elementu na schemacie.
       * @param y nowa odcieta
       * @see getX(); getY(); getRot(); setX(); setRot();
       */ 
      void setY(int y){m_y=y;}
      /**
       * Ustawia rzedna elementu na schemacie.
       * @param rot nowa rotacja, mozliwoci : prawo(0), gora(1), lewo(2),
       *        dol(3). Pozostale wartosc sa rownowazne z rotacja prawa.
       * @see getX(); getY(); getRot(); setX(); setY();
       */ 
      void setRot(int rot){m_rot=rot;}

      /**
       * Polacz pin do innego elementu.
       * @param src_pin pin obecnego elementu  
       * @param *dest wskaznik do elementu do ktorego ma byc nawiazane polaczenie
       * @param dest_pin pin w elemencie docelowym do ktorego  mabyc nnaae polaczenie
       * @see disconect(unsigned int src_pin);
       */ 
      bool conect(CElement *dest,unsigned int src_pin,unsigned int dest_pin);
      /**
       * Rozlacza pin.
       * @param src_pin pin ktory mamy odlaczyc 
       * @see conect(CElement *dest,unsigned int src_pin,unsigned int dest_pin);
       */ 
      bool disconect(unsigned int src_pin);
    
      /**
       * Funkcja wypisuje podstawowe parametry elementu,
       * storzona w celach testowych.
       */ 
      virtual void print(){ std::cout << m_symbol << ":"<< m_name << std::endl;}
      /**
       * Funkcja wypisuje polaczenia z innymi elementami,
       * storzona w celach testowych.
       */ 
      void print_conections();
    
      /**
       * Wektor ktory przechowywuje wskazniki do pinow,
       */ 
      std::vector<CPin *> pins;
      /**
       * WTF ? nie mam pomyslu do jakiej galezi to moze byc wskaznik ;/
       */ 
      CBranch *branch;    
      /**
       * Wskaznik do sygnalu na danym elemencie.
       */ 
      CSignal *signal;

      /**
       * Rysuje element.
       */ 
      virtual void draw();
      /**
       * Rysuje polaczenia danego elementu z sasiadami.
       */ 
      void drawConnections();
      /**
       * Wyswietla nazwe elementu.
       */ 
      virtual void drawName();
      /**
       * Wyswietla jego wartosc.
       * @see getValueString
       * @see getValue
       */ 
      virtual void drawValue();

      
      // TODO: change these functions to one that returns Glib::Point.
      /**
       * Zwraca wspolrzedna x pinu o podanym numerze.
       * @param pinNumber 0 dla przodu, 1 dla tylu elementu`
       * @return wspolrzedna x pinu
       */
      virtual int getPinX(int pinNumber);
      /**
       * Zwraca wspolrzedna y pinu o podanym numerze.
       * @param pinNumber 0 dla przodu, 1 dla tylu elementu`
       * @return wspolrzedna y pinu
       */
      virtual int getPinY(int pinNumber);
    
      /**
       * Ustawia obwod, do ktorego nalezy element.
       */ 
      void setParent(CCircuit* parent) {m_ParentCircuit = parent;};
      
    protected: 
      /**
       * Nazwa elementu
       */ 
      std::string m_name;
      /**
       * Symbol elementu
       */ 
      std::string m_symbol;
      /**
       * Symbol jednostki.
       * Dla opornika "Ohm", dla kondensatora "F", dla cewki "H", etc.
       */ 
      std::string m_unitSymbol;
      /**
       * Symbol wartosci elementu.
       * Dla opornika "R", dla kondensatora "C", dla cewki "L", etc.
       */ 
      std::string m_valueSymbol;
      /**
       * Rzedna elementu na schemacie
       */ 
      int m_x;
      /**
       * Odcieta elementu na schemacie
       */ 
      int m_y;
      /**
       * Rotacja elementu na schemacie.
       */ 
      int m_rot;
      /**
       * Zwraca liczbe pinow danego elementu, 
       * (narazie zawsze =2);
       */ 
      unsigned int m_max_pin(){return pins.size();}
      /**
       * Wskaznik do obwod, do ktorego nalezy element.
       */ 
      CCircuit* m_ParentCircuit; 
  };


  /**
   * CExNode jest klasa reprezentujaca wierzcholek rozszerzony.
   * Jako nieliczne elementy zazwyczaj maja > 2 piny ktore wychodza
   * bezposrednio do elementow.
   * Zasadniczo klasa nie implementuje nic nowego, jednak istotne jest wprowadznie
   * nowej klasy w celu latwiejszego rozwiazywania obwodow.
   */
  class CExNode : public CElement
  {
    protected: 
 
    public:
      /**
       * Konstruktor.
       * @param *dst_item wskanik do elementu docelowego.
       * @param dst_pin numer pinu w elemencie docelowym.
       * @see ~ExNode();
       */
      CExNode(std::string name,int x=0,int y=0) : CElement("EXNODE",name,x,y){};
      /**
       * Destruktor.
       * @param *dst_item wskanik do elementu docelowego.
       * @param dst_pin numer pinu w elemencie docelowym.
       * @see ExNode();
       */
      ~CExNode(){ }
      /**
       * Funkcja dodaje koleiny pin
       * @param *pin wskaznik do nowego pinu.
       * @see count_pins();
       */
      void add(CPin * pin){ pins.push_back(pin); }
      /**
       * Zwraca ilosc pinow danego wezla
       */
      int count_pins(){return pins.size();}
      /**
       * Zmienna pomocnicza slozaca do nazywania koleinych wezlow,
       * tylko w celach testowych
       */
      unsigned int nr;
  };



/**
 * Klasa reprezentuje wierzcholek.
 */
class CNode : public CElement
{
protected: 
    
public:
   /**
    * Konstruktor.
    * @param name nazwa .
    * @param x rzedna elementu na schemacie.
    * @param y odcieta elementu na schemacie.
    * @see ~CNode();
    */
    CNode(std::string name,int x=0,int y=0) : CElement("NODE",name,x,y){
         for(int i=0;i<4;i++)
            pins.push_back(new CPin());
    };
   /**
    * Destruktor.
    * @see CNode();
    */    
    ~CNode(){};

    
    void draw();
    void drawName();
    void drawValue();

    // TODO: change these functions to one that returns Glib::Point.
    int getPinX(int pinNumber) {return m_x;}
    int getPinY(int pinNumber) {return m_y;}

   /**
    * wskaznik do wierzcholka glownego, po zciagnieciu ukladu.
    */    
    CExNode *papa;
};


/**
 * Klasa reprezentuje elementy pasywne (R,L,C).
 */
class CElementPasive:public CElement
{
public:

    /**
     * Konstruktor bezargumentowy.
     * @see CElementPasive(std::string symbol,std::string name,int x=0,int y=0,int rot=0);~CElementPasive();     
     */
    CElementPasive():CElement(){};
    /**
     * Konstruktor.
     * @param symbol typ elementu.
     * @param name nazwa.
     * @param x rzedna elementu na schemacie.
     * @param y odcieta elementu na schemacie.
     * @param rot rotacja elementu na schemacie, mozliwoci : prawo(0), gora(1), lewo(2),
     *        dol(3). Pozostale wartosc sa rownowazne z rotacja prawa.
     * @see CElement();CElementPasive();~CElementPasive();
     */
    CElementPasive(std::string symbol,std::string name,int x=0,int y=0,int rot=0):
        CElement(symbol,name,x,y,rot){};
    
    /**
     * Wirtualny destruktor.
     * @see CElementPasive();
     */
    virtual ~CElementPasive(){};

    /**
     * Zwraca impedancje.
     * @param omega czestosc kolowa wymuszenia
     * @return wartosc impedancji
     * @see Complex getYc(double omega);
     */
    virtual Complex getZc(double omega){return Complex(0,0);};
    /**
     * Zwraca konduktancje.
     * @param omega czestosc kolowa wymuszenia
     * @return wartosc konduktancji
     * @see Complex getZc(double omega);
     */    
     virtual Complex getYc(double omega){return Complex(0,0);};
    
};
/**
 * Klasa reprezentuje elementy aktywne.
 */
class CElementActive:public CElement
{
protected:
    //! czestosc kolowa elementu
    double m_omega;
    //! faza poczatkowa. 
    double m_fi0;  
public:
    /**
     * Konstruktor bezargumentowy.
     * @see CElementActive(std::string symbol,std::string name,double omega=0,double fi0=0,int x=0,int y=0,int rot=0);~CElementActive();       
     */
    CElementActive():CElement(){};
    /**
     * Konstruktor.
     * @param symbol typ elementu.
     * @param name nazwa.
     * @param x rzedna elementu na schemacie.
     * @param y odcieta elementu na schemacie.
     * @param omega czestosc kolowa wymuszenia.
     * @param fi0 faza poczatkowa.
     * @param rot rotacja elementu na schemacie, mozliwoci : prawo(0), gora(1), lewo(2),
     *        dol(3). Pozostale wartosc sa rownowazne z rotacja prawa.
     * @see CElement();CElementActive();~CElementActive();
     */
    
    CElementActive(std::string symbol,std::string name,double omega=0,double fi0=0,int x=0,int y=0,int rot=0):
      CElement(symbol,name,x,y,rot)
      {
        m_omega=omega;
        m_fi0=fi0;
      };
    
    /**
     * Wirtualny destruktor.
     * @see CElementPasive();
     */
    virtual ~CElementActive(){};

    /**
     * Zwraca czestoc kolowa.
     * @return czestoc kolowa
     * @see setOmega(double omega);
     */
    double getOmega(void){return m_omega;};
    /**
     * Ustawia czestoc kolowa.
     * @param czestoc czestosc kolowa
     * @see getOmega(double omega);
     */    
    void setOmega(double omega){m_omega=omega;};
    /**
     * Zwraca faze poczatkowa.
     * @return faza poczatkowa
     * @see setFi0(double fi);
     */
    double getFi0(void){return m_fi0;};
    /**
     * Ustawia faze poczatkowa.
     * @param czestoc faza poczatkowa
     * @see getFi0(void);
     */    
    void setFi0(double fi){m_fi0=fi;};
  
};

/**
 * Klasa reprezentuje wierzcholek.
 */
class CElementControl:public CElement
{
protected:
    double m_gm;
    std::string m_on;
public:
    //! Konstruktor
    CElementControl():CElement(){};
    CElementControl(std::string symbol,std::string name,double gm, std::string on,int x=0,int y=0,int rot=0):
      CElement(symbol,name,x,y,rot)
      {
        m_gm=gm;
        m_on=on;
      };
    
    //! Destruktor
    virtual ~CElementControl(){};

    /**
     * Ustawia wspulczynnik wzmocnienia (transkonduktancji).
     * @return wspulczynnik wzmocnienia
     * @see setGm(double gm)(void);
     */  
    double getGm(void){return m_gm;};
    /**
     * Ustawia wspulczynnik wzmocnienia (transkonduktancji).
     * @param fm wspolczynik wzmoanienia
     * @see double getGm(void);
     */     
    void setGm(double gm){m_gm=gm;};
    
    /**
     * Zwraca element wzgledem ktorego ma sie odbywac sterowanie.
     * @return nazwa elementu odniesienia
     * @see setOn(std::string on);
     */     
    std::string getOn(void ){return m_on;}
    /**
     * Ustawia element wzgledem ktorego ma sie odbywac sterowanie.
     * @param on nazwa elementu odniesienia
     * @see getOn(void );
     */     
    void setOn(std::string on){m_on=on;};
    std::string getValueString();    
};

#endif
