/*
*  Parser.cc
*
*  Created by Marc Parizeau on Sat Oct 26 2002.
*  Copyright  2002 LVSN. All rights reserved.
*
*/

#include <xml/Parser.hh>
#include <stdexcept>
#include <fstream>
#include <sstream>

using namespace std;
using namespace XML;

Parser::Node::~Node(void)
{
	// delete childs
	if(mChild) delete mChild;
	mChild = 0;
	// delete siblings (without sibling recursion)
    while(mSibling){
    	Node* lTmp = mSibling;
    	mSibling = mSibling->mSibling;
    	if(lTmp->mChild) delete lTmp->mChild;
    	lTmp->mChild = lTmp->mSibling = 0;
    	delete lTmp;
    }
}

string Parser::Node::getTagAttribute(const string& inName) const
{
	map<string,string>::const_iterator lAttr = mAttributes.find(inName);
	return (lAttr != mAttributes.end() ? (*lAttr).second : "");
}

Parser::Node* Parser::Node::parse(Tokenizer& inTokenizer)
{
	Node* lNode = 0;
	// look for start tag
	string lToken;
	inTokenizer.setWhiteSpace("");
	inTokenizer.setSingleCharTokens("<");
	if(!inTokenizer.getNextToken(lToken)) return false;
	if(lToken.find_first_not_of(" \t\n\r") == string::npos) {
		// flush any leading white space
		if(!inTokenizer.getNextToken(lToken)) return 0;
	}
	
	if(lToken == "<") {
		// check for end tag
		char lChar;
		inTokenizer.getStreamPtr()->get(lChar);
		if(lChar == '/') {
			// found end tag; 
			return 0;
		}
		inTokenizer.getStreamPtr()->putback(lChar);
		// found start tag
		lNode = new Node;
		lNode->parseStartTag(inTokenizer, lToken);
		if(lToken == "/") {
			// found empty tag; next token must be '>'
			inTokenizer.setWhiteSpace("");
			inTokenizer.setSingleCharTokens(">");
			if(!inTokenizer.getNextToken(lToken)) lNode->throwError(inTokenizer, "unexpected eof");
			if(lToken != ">") lNode->throwError(inTokenizer, "invalid start tag");
		} 
		else 
			if(lNode->getType() == eTag) {
				// parse all child
				Node* lChild;
				Node* lLastChild = 0;
				while(lChild = parse(inTokenizer)){
					// add new sibling
					if(lLastChild == 0) 
						lLastChild = lNode->mChild = lChild;
					else{
						lLastChild->mSibling = lChild;
						lLastChild = lChild;
					}
				}
				// test for valid end tag
				inTokenizer.setWhiteSpace("");
				inTokenizer.setSingleCharTokens(" \t\n\r>");
				if(!inTokenizer.getNextToken(lToken)) lNode->throwError(inTokenizer, "unexpected eof");
				if(lToken != lNode->mValue) lNode->throwError(inTokenizer, "invalid end tag");
				// next token must be '>'
				inTokenizer.setWhiteSpace(" \t\n\r");
				inTokenizer.setSingleCharTokens(">");
				if(!inTokenizer.getNextToken(lToken)) lNode->throwError(inTokenizer, "unexpected eof");
				if(lToken != ">") lNode->throwError(inTokenizer, "invalid end tag");
			}
	} 
	else {
		// found a string node
		lNode = new Node;
		lNode->mType = eString;
		lNode->mValue = lToken;
	}
	return lNode;
}

void Parser::Node::parseAttributeList(Tokenizer& inTokenizer, string& outToken)
{
	inTokenizer.setWhiteSpace(" \t\n\r");
	inTokenizer.setSingleCharTokens("=/>");
	// next token should be an attribute name
	if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
	// parse all attributes
	while(outToken != "/" && outToken != ">"){
		// ok, found an attribute name!
		string lName = outToken;
		// next token should be '='
		inTokenizer.setSingleCharTokens("=");
		if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
		if(outToken != "=") throwError(inTokenizer, "invalid attribute");
		inTokenizer.setSingleCharTokens("\"");
		// next token must be '"'
		if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
		if(outToken != "\"") throwError(inTokenizer, "invalid attribute value");
		// next token is either a value or '"'
		inTokenizer.setWhiteSpace("");
		inTokenizer.setSingleCharTokens("\"");
		if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
		if(outToken == "\"") {
			// value is empty string
			mAttributes[lName] = "";
		} 
		else {
			mAttributes[lName] = outToken;
			inTokenizer.getNextToken(outToken);
			if(outToken != "\"") throwError(inTokenizer, "unexpected eof");
		}
		
		inTokenizer.setWhiteSpace(" \t\n\r");
		inTokenizer.setSingleCharTokens("=/>");
		if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
	}
}

void Parser::Node::parseStartTag(Tokenizer& inTokenizer, string& outToken)
{
	// parse tag name
	inTokenizer.setWhiteSpace("");
	inTokenizer.setSingleCharTokens(" \t\n\r/>");
	if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
	if(outToken.find_first_of(" \t\n\r/>") != string::npos) throwError(inTokenizer, "invalid start tag name");
	switch(outToken[0]) {
		case '?': case '!':
				if(outToken[0] == '?') mType = eDecl;
				else mType = eSpecial;
				// parse special tag
				mValue = outToken;
				inTokenizer.setSingleCharTokens(">");
				if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
				if(outToken != ">") {
					mValue += outToken;
					if(!inTokenizer.getNextToken(outToken)) throwError(inTokenizer, "unexpected eof");
				}
				break;
		default:
				// token is regular tag name
				mType = eTag;
				mValue = outToken;
				parseAttributeList(inTokenizer, outToken);
	}
}

void Parser::Node::serialize(Streamer& outStream, bool inIndent) const
{
	bool lLastChildIsTag = true;
	switch(mType){
		case eDecl: case eSpecial:
				outStream.insertSpecialTag(getValue());
				break;
		case eTag:
				outStream.openTag(getTagName(), inIndent);
				// serialize attribute list
				for(map<string,string>::const_iterator i = mAttributes.begin(); i != mAttributes.end(); ++i){
					outStream.insertAttribute((*i).first, (*i).second);
				}
				// serialize child nodes
				for(Node* lChild = mChild; lChild != 0; lChild = lChild->mSibling){
					lChild->serialize(outStream, lLastChildIsTag);
					lLastChildIsTag = (lChild->getType() != eString);
				}
				// indent closing tag only if last child is a tag
				outStream.closeTag(lLastChildIsTag);
				break;
		case eString:
				outStream << getValue();
				break;
		default:
				throw runtime_error("Internal error: unknown node type");
	}
}

void Parser::Node::throwError(Tokenizer& inTokenizer, const string& inMessage) const
{
	ostringstream lStream;
	lStream << "\n";
	if(inTokenizer.getStreamName() != "") lStream << "XML::Parser:: In \"" << inTokenizer.getStreamName() << "\", line ";
	else lStream << "Line ";
	lStream << inTokenizer.getLineNumber() << "\n" << mValue << ": " << inMessage;
	throw runtime_error(lStream.str());
}

void Parser::parse(const std::string& inFileName)
{
	ifstream lStream(inFileName.c_str());
	if(!lStream.good()) throw runtime_error(string("<Parser::parse> unable to open file ")+inFileName);
	parse(lStream, inFileName);
}

void Parser::parse(std::istream& inStream, const std::string& inName)
{
	Tokenizer lTokenizer(inStream);
	lTokenizer.setStreamName(inName);
	Node* lNode = 0;
	Node* lLastSibling = 0;
	while(lNode = Node::parse(lTokenizer)){
		if(lLastSibling == 0) lLastSibling = mRoot = lNode;
		else{
			lLastSibling->mSibling = lNode;
			lLastSibling = lNode;
		}
	}
	if(!inStream.eof()) lNode->throwError(lTokenizer, "invalid xml tag");
}
