/*
 * RolLerCoaster Linear Networks Solver
 *
 * RolLerCoaster is the legal property of its developers, whose names are 
 * listed in the COPYRIGHT file distributed with this source distribution.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the free software foundation; either version 2 of the license, or
 * (at your option) any later version.
 *
 * this program is distributed in the hope that it will be useful,
 * but without any warranty; without even the implied warranty of
 * merchantability or fitness for a particular purpose.  see the
 * gnu general public license for more details.
 *
 * you should have received a copy of the gnu general public license
 * along with this program; if not, write to the free software
 * foundation, inc., 59 temple place, suite 330, boston, ma  02111-1307  usa
 *
 */


#include <iostream>
#include <gtkmm/window.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/stock.h>

#include <xml/Parser.hh>
#include "networksarea.hh"
#include "mainwindow.hh"
#include <CRlcException.hh>
#include <CCircuitSolver.hh>

using namespace RLC;

CNetworksArea::CNetworksArea() 
{
  // Setting requested drawing area size
  set_size_request(700,500);

  // Connecting events with signal handlers
  signal_expose_event().connect( sigc::mem_fun(*this,
	&CNetworksArea::on_networksarea_expose_event) );
  signal_configure_event().connect( sigc::mem_fun(*this,
	&CNetworksArea::on_networksarea_configure_event) );
  signal_motion_notify_event().connect( sigc::mem_fun(*this,
	&CNetworksArea::on_networksarea_motion_notify_event) );
  signal_button_press_event().connect( sigc::mem_fun(*this,
	&CNetworksArea::on_networksarea_button_press_event) );
  signal_button_release_event().connect( sigc::mem_fun(*this,
	&CNetworksArea::on_networksarea_button_release_event) );

  // Setting colors that will be used.
  m_Color_Blue  = Gdk::Color("blue");
  m_Color_Green = Gdk::Color("green");
  m_Color_Red   = Gdk::Color("red");
  m_Color_Black = Gdk::Color("black");
  m_Color_White = Gdk::Color("white");
  m_Color_Beige = Gdk::Color("cornsilk1");

  // Setting colors for every "element" to draw. In future this will be
  // read from config file. @FUTURE_CONFIG@

  m_Color_BG = m_Color_Black;
  m_Color_FG = m_Color_White;
  m_Color_Element = m_Color_Green;
  m_Color_Node   = m_Color_Blue;
  m_Color_Branch = m_Color_Red;
  m_Color_NameText = m_Color_White;
  m_Color_ValueText = m_Color_White;
  m_Color_FrameBG = m_Color_Beige;
  m_Color_Chart   = m_Color_Red;
  m_Color_Axis    = m_Color_Black;
  
  // We can only allocate colormap here, so we do this. The rest will happen
  // in on_configure event handler. 
  Glib::RefPtr<Gdk::Colormap> colormap = get_default_colormap();

  colormap->alloc_color(m_Color_BG);
  colormap->alloc_color(m_Color_FG);
  colormap->alloc_color(m_Color_Element);
  colormap->alloc_color(m_Color_Node);
  colormap->alloc_color(m_Color_Branch);
  colormap->alloc_color(m_Color_NameText);
  colormap->alloc_color(m_Color_ValueText);
  colormap->alloc_color(m_Color_FrameBG);
  colormap->alloc_color(m_Color_Chart);
  colormap->alloc_color(m_Color_Axis);

  // Setting size o a square, that will be an elemental field to use. In future
  // for config file. @FUTURE_CONFIG@
  m_SquareSize = 30;

  // Setting size of a node. @FUTURE_CONFIG@
  m_NodeSize = static_cast<int>(0.15*m_SquareSize);

  // Adding additional masks that drawing area do not normally use.
  add_events(Gdk::EXPOSURE_MASK |
             Gdk::POINTER_MOTION_MASK |
	     Gdk::POINTER_MOTION_HINT_MASK);

  // Setting Circuit parent to this object. So that CCircuit class object
  // knows were to draw itself.
  m_Circuit_Edited.setParent(this);
  // CCircuitSolver ;
	                      
  // This is done to draw the circuit for the first time and configure
  // pixmap where it is stored.
  m_circuitIsChanged   = true;
  m_pixmapIsConfigured = false;
  m_circuitIsLoaded = false; 
  m_circuitIsSolved = false;

  // Creating new Pango::Context for this widget (CNetworksArea),
  // for further use in text rendering methods (for example 
  // names of elements).
  m_RefContext_Main = create_pango_context();

  // Setting up signal frame object. Especially it's parent
  // to this CNetworksArea object.
  m_SignalFrame_Element.setParent(this);
  
  // Initializing a pointer to parent CMainWindow widget.
  // m_parent = 0;
  
  // m_maxFieldX = 200;
  // m_maxFieldY = 100;
  // mf_initFieldElement();
  
  std::cout << "CNetworksArea::CNetworksArea called." << std::endl;

}


CNetworksArea::~CNetworksArea()
{

  std::cout << "CNetworksArea::~CNetworksArea called." << std::endl;
}


void CNetworksArea::openCircuitFile()
{
  Gtk::FileChooserDialog dialog("Please choose a network file", Gtk::FILE_CHOOSER_ACTION_OPEN);
  // TODO: do what is written below. Use RefPtr<Gtk::Window&> instead of CMainWindow* ??
  //dialog.set_transient_for(m_parent); // Setting this so that when open dialog is shown
                                      // the user cannot manipulate in main window.

  // Adding response buttons for dialog:
  dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);
  
  // Adding filters (also for *.rlc files):
  Gtk::FileFilter filter_any;
  filter_any.set_name("Any files");
  filter_any.add_pattern("*");
  dialog.add_filter(filter_any);

  Gtk::FileFilter filter_rlc;
  filter_rlc.set_name("RLC files");
  filter_rlc.add_pattern("*.rlc");
  dialog.add_filter(filter_rlc);

  // Showing dialog and wainting for result to analyze it.
  int result = dialog.run();

  // TODO: do sth with brackets struct below. Too deep indents??
  // Handling the response:
  switch (result) {
    case(Gtk::RESPONSE_OK):
      {
	std::string fileName = dialog.get_filename(); // Change it to Glib::ustring??

	try{
	  // TODO: check whether it is the same file. Or maybe do not check this?!
	  std::cout << "x" << std::endl;
	  clearCircuit();
	  std::cout << "x" << std::endl;
	  m_Circuit_Edited.load(fileName); 
	  std::cout << "x" << std::endl;
	  m_circuitKiller.load(&m_Circuit_Edited);
	  std::cout << "x" << std::endl;
	  m_circuitIsChanged = true;
	  mf_drawCircuit();
	  queue_draw();
	  m_circuitIsLoaded = true;
	  // on_networksarea_expose_event(); // Doing this to reload view. 
	}
	catch(CRlcException& e){
	  std::cerr << e.what() << std::endl;

	  Gtk::MessageDialog errorDialog(e.what(),false, Gtk::MESSAGE_ERROR);
	  errorDialog.run();
	}

	break;
      }
    case (Gtk::RESPONSE_CANCEL):
      {
	std::cout << "Cancel clicked." << std::endl;
	break;
      }
    default:
      {
	std::cout << "Unhandled button click." << std::endl;
	break;
      }
  }
}


void CNetworksArea::solveCircuit()
{

  if (m_circuitIsLoaded) {
    if (!m_circuitIsSolved) {
      try {
	m_circuitKiller.solve();

	m_circuitIsSolved = true;
      }
      catch (CRlcException& e) {
	std::cerr << e.what() << std::endl;

	Gtk::MessageDialog errorDialog(e.what(),false, Gtk::MESSAGE_ERROR);
	errorDialog.run();
      }
    }
    else {
      std::string msg = "This circuit is already solved.";
      std:: cout << msg << std::endl;

      Gtk::MessageDialog errorDialog(msg,false, Gtk::MESSAGE_WARNING);
      errorDialog.run();
    }
  }
  else {
    std::string msg = "Create or load a circuit file first.";
    std:: cout << msg << std::endl;

    Gtk::MessageDialog errorDialog(msg,false, Gtk::MESSAGE_WARNING);
    errorDialog.run();
  }
}


void CNetworksArea::clearCircuit()
{
  // try {
    // m_circuitKiller.clear();
    // m_circuitIsSolved = false;
  // }
  // catch (CRlcException& e) {
    // std::cerr << e.what() << std::endl;
// 
    // Gtk::MessageDialog errorDialog(e.what(),false, Gtk::MESSAGE_ERROR);
    // errorDialog.run();
  // }

  try {
    m_Circuit_Edited.clear();
    m_circuitIsLoaded = false;
  }
  catch (CRlcException& e) {
    std::cerr << e.what() << std::endl;

    Gtk::MessageDialog errorDialog(e.what(),false, Gtk::MESSAGE_ERROR);
    errorDialog.run();
  }

  m_circuitIsChanged = true;
}


// void CNetworksArea::setParent(CMainWindow* parent)
// {
  // m_parent = parent;
// }


bool CNetworksArea::mf_drawNetting()
{
  Glib::RefPtr<Gdk::Window> window = get_window();
  int width  = get_width();
  int height = get_height();

  for(int y=0; y < height; y+=m_SquareSize)
    for(int x=0; x < width; x+=m_SquareSize)
      m_RefPixmap_NetworksArea->draw_point(m_RefGC_NetworksArea,x,y);

  
  std::cout << "CNetworksArea::mf_drawNetting called." << std::endl;
  
  return true;
}


bool CNetworksArea::mf_drawArrow(int fieldX, int fieldY, int rot, float proportion)
{

  //TODO: change this below to not use 'case'. Figure out some simple math formula for this.
  // Setting variables based on rotatation for futher use.
  int rot_x = 0, rot_y = 0;
  switch (rot) {
    case  1: rot_y = 1; break;
    case  2: rot_x =-1; break;
    case  3: rot_y =-1; break;
    default: rot_x = 1; 
  }

  // This array has the meaning as below:
  //    0  1
  // 0 [x][y] - arrow front
  // 1 [x][y] - arrow back
  int arrowCoords[2][2];
  float arrowProportion = proportion;
  
  arrowCoords[0][0] = static_cast<int>( (fieldX+1+arrowProportion*rot_x)*m_SquareSize );
  arrowCoords[0][1] = static_cast<int>( (fieldY+1-arrowProportion*rot_y)*m_SquareSize );
  arrowCoords[1][0] = static_cast<int>( (fieldX+1-arrowProportion*rot_x)*m_SquareSize );
  arrowCoords[1][1] = static_cast<int>( (fieldY+1+arrowProportion*rot_y)*m_SquareSize );
 
  // TODO: Is there a simplier call for drawing a line? I.e. using array?
  m_RefPixmap_NetworksArea->draw_line(
      m_RefGC_NetworksArea,
      arrowCoords[0][0],
      arrowCoords[0][1],
      arrowCoords[1][0],
      arrowCoords[1][1]);
  
  int edgeCoords[2][2];
  float edgeProportion = 0.5*arrowProportion;
  edgeCoords[0][0] = static_cast<int>( arrowCoords[0][0]-(rot_x*edgeProportion+rot_y*edgeProportion)*m_SquareSize+0.5 );
  edgeCoords[0][1] = static_cast<int>( arrowCoords[0][1]-(rot_x*edgeProportion-rot_y*edgeProportion)*m_SquareSize+0.5 );
  edgeCoords[1][0] = static_cast<int>( arrowCoords[0][0]-(rot_x*edgeProportion-rot_y*edgeProportion)*m_SquareSize+0.5 );
  edgeCoords[1][1] = static_cast<int>( arrowCoords[0][1]+(rot_x*edgeProportion+rot_y*edgeProportion)*m_SquareSize+0.5 );
  
  // TODO: The same as above.
  m_RefPixmap_NetworksArea->draw_line(
      m_RefGC_NetworksArea,
      arrowCoords[0][0],
      arrowCoords[0][1],
      edgeCoords[0][0],
      edgeCoords[0][1]);
  
  m_RefPixmap_NetworksArea->draw_line(
      m_RefGC_NetworksArea,
      arrowCoords[0][0],
      arrowCoords[0][1],
      edgeCoords[1][0],
      edgeCoords[1][1]);

  // std::cout << "CNetworksArea::mf_DrawArrow(" << fieldX <<"," << fieldY << "," << rot << ")  called ";
  // std::cout << "|| Arrow Coords: [" << arrowCoords[0][0] << "," << arrowCoords[0][1] << "," 
            // << arrowCoords[1][0] << "," << arrowCoords[1][1] << "] ";
  // std::cout << "|| Edge Coords: [" << edgeCoords[0][0] << "," << edgeCoords[0][1] << "," 
            // << edgeCoords[1][0] << "," << edgeCoords[1][1] << "] " << std::endl;
}


Gdk::Point CNetworksArea::mf_fieldCoords(Gdk::Point coords)
{
  return Gdk::Point(static_cast<int>( coords.get_x()/m_SquareSize ),
                    static_cast<int>( coords.get_y()/m_SquareSize ));
}


CElement* CNetworksArea::mf_getFieldElement(Gdk::Point fieldCoords)
{
   for(int i=0;i<m_Circuit_Edited.items.size();++i) {
     CElement* pElement = m_Circuit_Edited.items[i];
     // std::cout << "Singal [" << i << "]=" << pElement->signal << std::endl;
     if ( Gdk::Point(pElement->getX(), pElement->getY()) == fieldCoords ) {
       // std::cout << "Returning element number: " << i << std::endl;
       return pElement;
     }
   }

  return 0;
}


void CNetworksArea::mf_drawCircuit()
{
  Glib::RefPtr<Gdk::Window> window = get_window();
  int width  = get_width();
  int height = get_height();
  m_RefPixmap_NetworksArea = Gdk::Pixmap::create(
      window, 
      width,
      height,
      -1);

  if (m_circuitIsChanged) {
    mf_drawNetting();
    m_Circuit_Edited.draw();
    // This will be uncomment when double buffers will be used.
    // m_circuitIsChanged = false;
  }

}



// bool CNetworksArea::mf_initFieldElement()
// {
 // for(int x=0; x<m_maxFieldX; ++x) 
   // for (int y=0; y<m_maxFieldY; ++y) {
     // m_vFieldElement[x][y] = 0;
   // }
 // 
 // return true;
// }


bool CNetworksArea::on_networksarea_configure_event(GdkEventConfigure* event)
{
  // TODO: insert here a double Pixmap handling: one for cicuit, and 
  // one for window drawing. The idea is to draw to cicruit pixmap,
  // and copy to window pixmap what is really in a window. 
  // if (!m_pixmapIsConfigured) {
  // Glib::RefPtr<Gdk::Screen> defaultScreen = Gdk::Screen::get_default();
  // m_pixmapIsConfigured = true;
  // }

  
  Glib::RefPtr<Gdk::Window> window = get_window();
  int width  = get_width();
  int height = get_height();

  m_RefGC_NetworksArea = Gdk::GC::create(window);

  window->set_background(m_Color_BG);
  // window->clear(); // DO WE HAVE TO CLEAR WINDOW REALLY??
  m_RefGC_NetworksArea->set_foreground(m_Color_FG);
  //m_RefGC_NetworksArea->set_background(m_Color_BG);

  m_RefPixmap_NetworksArea = Gdk::Pixmap::create(
      window, 
      width,
      height,
      -1);

  mf_drawCircuit();

  std::cout << "Width: " << width << "  Height: " << height << "  Window: " << window << std::endl;

  m_circuitIsChanged = true;

  std::cout << "CNetworksArea::on_networksarea_configure_event called." << std::endl;

  return true;
}


bool CNetworksArea::on_networksarea_expose_event(GdkEventExpose* event)
{

  get_window()->draw_drawable(
      m_RefGC_NetworksArea,
      m_RefPixmap_NetworksArea,
      event->area.x, event->area.y,
      event->area.x, event->area.y,
      event->area.width, event->area.height);

  std::cout << "CNetworksArea::on_networksarea_expose_event called." << std::endl;

  return false;
}


bool CNetworksArea::on_networksarea_motion_notify_event(GdkEventMotion* event)
{
  // Present under-pointer field coords used to check if the pointer left the field.
  // If yes we will have to check if it left the element area, if that is positive
  // too we will have to draw a new signal characteristic, if any. This is used 
  // to lower the drawing of already drawn characteristics frames.
  static Gdk::Point s_pfCoords(0,0);
  static CElement* s_foundElement = 0;
  
  if(event && event->window)
  {
    const Glib::RefPtr<Gdk::Window> refWindow = 
      Glib::wrap((GdkWindowObject*) event->window, true); /* true = take copy */
    
    if(refWindow)
    {
      int x = 0, y = 0;
      Gdk::ModifierType state = Gdk::ModifierType(0);

      refWindow->get_pointer(x, y, state);

      // coords of field that is below pointer
      Gdk::Point pfCoords( mf_fieldCoords( Gdk::Point(x,y) ) ); 

      if (s_pfCoords != pfCoords) {
	CElement* foundElement = mf_getFieldElement(pfCoords);
	if (foundElement != s_foundElement) {
	  mf_drawCircuit();
	  if (foundElement != 0) {
	    CSignal* foundSignal = foundElement->signal;
	    if ( (foundSignal != 0) /*&& (foundElement->getType() == "NODE")*/ ) {
	      std::cout << "Found element != 0" << std::endl;
	      m_SignalFrame_Element.setSignal(foundSignal);
	      std::cout << "Set the signal in signal frame." << std::endl;
	      m_SignalFrame_Element.setCoords( Gdk::Point(
		                                pfCoords.get_x()*m_SquareSize,
		                                pfCoords.get_y()*m_SquareSize ));
	      std::cout << "Set the signal frame coords. " << std::endl;
	      m_SignalFrame_Element.draw();
	    }

	  }

	  queue_draw();

	  s_foundElement = foundElement;
	}
	// std::cout << "Under-pointer field coords: [" << pfCoords.get_x() 
	          // << "," << pfCoords.get_y() << "]" << std::endl;
	

	s_pfCoords = pfCoords;
      }
    }
  }
  

  // std::cout << "CNetworksArea::on_networksarea_motion_notify_event called." << std::endl;

  // Event handled so stopping porcessing
  return true;
}


bool CNetworksArea::on_networksarea_button_press_event(GdkEventButton* event)
{

  std::cout << "CNetworksArea::on_networksarea_button_press_event called." << std::endl;
}


bool CNetworksArea::on_networksarea_button_release_event(GdkEventButton* event)
{

  std::cout << "CNetworksArea::on_networksarea_button_release_event called." << std::endl;
}

