/*
 * RolLerCoaster Linear Networks Solver
 *
 * RolLerCoaster is the legal property of its developers, whose names are 
 * listed in the COPYRIGHT file distributed with this source distribution.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the free software foundation; either version 2 of the license, or
 * (at your option) any later version.
 *
 * this program is distributed in the hope that it will be useful,
 * but without any warranty; without even the implied warranty of
 * merchantability or fitness for a particular purpose.  see the
 * gnu general public license for more details.
 *
 * you should have received a copy of the gnu general public license
 * along with this program; if not, write to the free software
 * foundation, inc., 59 temple place, suite 330, boston, ma  02111-1307  usa
 *
 */
 
#ifndef NETWORKSAREA_HH
#define NETWORKSAREA_HH

#include <gtkmm/drawingarea.h>
#include <gdkmm/colormap.h>
#include <pangomm/context.h>

#include <CCircuit.hh>
#include <signalframe.hh>
#include <CCircuitSolver.hh>

/**
 * Main RolLerCoaster namespace.
 * Namespace that consists of every class that is made for R(ol)L(er)C(oaster).
 */
namespace RLC
{
  // class CMainWindow;

  /**
   * A circuit drawing class.
   * This class handles everything that is connected with drawing. 
   * It also handles opening the circuit to draw.
   */
  class CNetworksArea : public Gtk::DrawingArea
  {
    friend class CCircuit;
    friend class CElement; 
    friend class CVoltage;
    friend class CVoltage_u;
    friend class CVoltage_i;
    friend class CCurent;
    friend class CCurent_u;
    friend class CCurent_i;
    friend class CResistor;
    friend class CCapasistor;
    friend class CInductor;
    friend class CNode;
    friend class CSignalFrame;

    public:
      /**
       * A constructor.
       * Reads config, set variables, connect singals with handlers, allocates
       * color maps.
       */
      CNetworksArea();

      /**
       * A destructor.
       * Does not really do anything for now.
       */
      ~CNetworksArea();

      /**
       * Circuit opening method.
       * Calls file chooser dialog, to give user a way to select circuit file. Then
       * loads it, and draw.
       */
      void openCircuitFile();

      /**
       * Circuit solving method.
       * After calling this a m_Circuit_Edited will be solved by m_circuitKiller.
       * @see m_Circuit_Edited
       * @see m_circuitKiller
       */
      void solveCircuit();

      /**
       * Circuit clearing method.
       * This calls clearing methods for m_Circuit_Edited and m_circuitKiller in
       * a secure way of opening another file for example.
       */
      void clearCircuit();


      // void setParent(CMainWindow* parent); 

    protected:
      /**
       * Draws netting.
       * Draws the basic netting that is at a background. The field size it defined
       * in m_squareSize variable.
       * @see m_squareSize
       */
      bool mf_drawNetting();
      
      /**
       * Draws arrow.
       * This method is used to draw arrows oriented in 4 possible ways: right, left, 
       * up, down. This is used to draw elements with arrow, for example voltages or
       * currents.
       * @param fieldX a horizontal coordinate of a netting field (most left are zeros).
       * @param fieldY a vertical coordinate of a netting field (most up are zeros).
       * @param rot rotation of an arrow. Possibilities are: right(0), up(1), left(2),
       *        down(3). All other values are threatened as zero.
       * @param proportion a proportion of an arrow. This value is understood as the part
       *        of squareSize that will be the half of arrow's length.
       * @see m_SquareSize
       */ 
      bool mf_drawArrow(int fieldX, int fieldY, int rot = 0, float proportion = 0.6);

      /**
       * Function returning network field coords.
       * This function converts normal pixel coords to network field coords.
       * @param coords pixel coordinates on the NetworkArea
       * @return coordinates of field where pixel with coords from argument is placed.
       */
      Gdk::Point mf_fieldCoords(Gdk::Point coords);

      /**
       * Return a pointer to element in a field.
       * This function returns a pointer to CElement object that is placed in a field 
       * that coords are given.
       * @param fieldCoords coords of a field from which we want to get a CElement pointer
       * @return return CElement pointer to a CElement that is a fieldCoords field. Or 0
       *         if there is none.
       */
      CElement* mf_getFieldElement(Gdk::Point fieldCoords);

      /**
       * Circuit drawing method.
       * This method draws a network netting and a circuit itself. It clears a 
       * m_RefPixmap_NetworksArea so that the circuit is not drawn one on another/
       * In future this will be only called when the circuit will really change.
       */
      void mf_drawCircuit();
      
      // bool mf_initFieldElement();
      
      bool on_networksarea_configure_event(GdkEventConfigure* event);
      bool on_networksarea_expose_event(GdkEventExpose* event);
      bool on_networksarea_motion_notify_event(GdkEventMotion* event);
      bool on_networksarea_button_press_event(GdkEventButton* event);
      bool on_networksarea_button_release_event(GdkEventButton* event);
      
      /**
       * A pixmap where the circuit is drawn.
       */
      Glib::RefPtr<Gdk::Pixmap> m_RefPixmap_NetworksArea;
      /**
       * Main gtkmm graphics context used as argument in all drawing functions.
       */
      Glib::RefPtr<Gdk::GC> m_RefGC_NetworksArea;
      
      // TODO: rewrite this in some kind of array.
      /**
       * Color variable.
       * Used so that we don's have to reallocate it every time we use it. Other color
       * variables are the same.
       */
      Gdk::Color m_Color_Blue, m_Color_Green, m_Color_Red, m_Color_Black, m_Color_White,
	         m_Color_Beige;
      Gdk::Color m_Color_BG, m_Color_FG, m_Color_Element, m_Color_Node, m_Color_Branch,
	         m_Color_NameText, m_Color_ValueText, m_Color_FrameBG, m_Color_Chart,
		 m_Color_Axis;
      
      /**
       * Netting field size.
       * This is the length of a side of the square (field) that is the background
       * for drawing circuits.
       */
      int m_SquareSize;
      /**
       * Node size.
       * This is a node size defined as 0.15 of m_SquareSize so that a proporction 
       * is good (const know) after resizing a network.
       */
      int m_NodeSize;
      
      /**
       * Edited Circuit.
       */
      CCircuit m_Circuit_Edited;
      /**
       * CCircuitSolver variable used to solver m_Circuit_Edited.
       */
      CCircuitSolver m_circuitKiller;
      /**
       * Signal frame variable.
       * This represents a frame with a signal that shows up when a mouse is over 
       * a element after solving the circuit.
       */
      RLC::CSignalFrame m_SignalFrame_Element;

      /**
       * If true the circuit is changed.
       * The same as some variables below this is a state variable used mostly
       * for drawing and keeping circuit states.
       */
      bool m_circuitIsChanged;
      /**
       * If true the pixmap is configured.
       */
      bool m_pixmapIsConfigured;
      /**
       * If true the circuit is loaded.
       */
      bool m_circuitIsLoaded;
      /**
       * If true the circuit is solved.
       */
      bool m_circuitIsSolved;

      /** 
       * Pango context mostly used for drawing Layout objects.
       */
      Glib::RefPtr<Pango::Context> m_RefContext_Main;

      // int m_maxFieldX;
      // int m_maxFieldY;
// 
      // std::vector<std::vector<CElement*> > m_vFieldElement;

      // CMainWindow* m_parent;
  };
  
}


#endif // NETWORKSAREA_HH

